from typing import List

from .dataclasses import Query
from .exceptions import LimsError
from .queries import EXP_WKFT_QUERY, IMAGE_WKFT_QUERY, PROBE_WKFT_QUERY

EXP_FILE_TYPES = [
    "NewstepConfiguration",
    "EcephysReplayStimulus",
    "EcephysAreaClassifications",
    "SideDlcOutputFile",
    "EyeTracking Ellipses",
    "RawEyeTrackingVideo",
    "RawEyeTrackingVideoMetadata",
    "EcephysSurgeryNotes",
    "RawFaceTrackingVideoMetadata",
    "RawBehaviorTrackingVideoMetadata",
    "StimulusPickle",
    "EyeDlcOutputFile",
    "OptoPickle",
    "MappingPickle",
    "RawBehaviorTrackingVideo",
    "FaceDlcOutputFile",
    "EcephysRigSync",
    "EcephysPlatformFile",
    "RawFaceTrackingVideo",
]

IMAGE_FILE_TYPES = [
    "EcephysPostStimulusRight",
    "EcephysPostInsertionLeft",
    "EcephysInsertionLocationImage",
    "EcephysPostInsertionRight",
    "EcephysPreExperimentRight",
    "EcephysPostExperimentLeft",
    "EcephysPreInsertionLeft",
    "EcephysFiducialImage",
    "EcephysPreInsertionRight",
    "EcephysBrainSurfaceRight",
    "EcephysPostExperimentRight",
    "EcephysBrainSurfaceLeft",
    "EcephysPostStimulusLeft",
    "EcephysPreExperimentLeft",
    "EcephysOverlayImage",
]

PROBE_FILE_TYPES = [
    "EcephysSortedParams",
    "EcephysSortedWhiteningMatInv",
    "EcephysSortedTemplates",
    "EcephysSortedSpikeClusters",
    "EcephysSortedChannelMap",
    "EcephysChannelStates",
    "EcephysSortedSpikeTimes",
    "EcephysSortedProbeInfo",
    "EcephysSortedChannelPositions",
    "EcephysTemplatesInd",
    "EcephysSortedSpikeTemplates",
    "EcephysSortedMeanWaveforms",
    "EcephysSortedEventTimestamps",
    "EcephysSortedSimilarTemplates",
    "EcephysSortedLfpTimestamps",
    "EcephysSortedMetrics",
    "EcephysProbeRawData",
    "EcephysSortedLfpContinuous",
    "EcephysSortedWhiteningMat",
    "EcephysSortedAmplitudes",
]

WKFT_NAMES = [*EXP_FILE_TYPES, *IMAGE_FILE_TYPES, *PROBE_FILE_TYPES]
"""Names of well known file types (wkft) on lims
"""


WKFT = {
    **(
        {
            name: {
                "template": EXP_WKFT_QUERY,
                "filters": [
                    (
                        "wkft",
                        name,
                    )
                ],
                "return_name": "wkf_path",
            }
            for name in EXP_FILE_TYPES
        }
    ),
    **(
        {
            name: {
                "template": IMAGE_WKFT_QUERY,
                "filters": [
                    (
                        "image_type",
                        name,
                    )
                ],
                "return_name": "image_path",
            }
            for name in IMAGE_FILE_TYPES
        }
    ),
    **(
        {
            name: {
                "template": PROBE_WKFT_QUERY,
                "filters": [
                    (
                        "wkft",
                        name,
                    )
                ],
                "return_name": "wkf_path",
            }
            for name in PROBE_FILE_TYPES
        }
    ),
}


def wkft_to_Query(wkft: str, exp_id: str) -> Query:
    """Transforms a well known file type (wkft) into a query.

    Args:
        wkft: A well known file type name
        exp_id: lims experiment id

    Returns:
        A resolved Query instance.

    Raises:
        LimsError: The supplied wkft name is unsupported.
    """
    try:
        wkft_dict = WKFT[wkft]
    except KeyError:
        raise LimsError("Unsupported wkft=%s" % wkft)

    return Query(
        query_str=wkft_dict["template"].format(exp_id),  # type: ignore[attr-defined]
        filters=wkft_dict["filters"],  # type: ignore[arg-type]
        return_name=wkft_dict["return_name"],  # type: ignore[arg-type]
    )


def get_wkft_names() -> List[str]:
    """Gets a list of all supported well known file types.

    Returns:
        List of supported well known file types.
    """
    return WKFT_NAMES



import os


def is_windows():
    """Detects whether the operating system is windows.

    Returns:
        windows detected
    """
    return os.name == "nt"


def fix_lims_path(path):
    """Fixes a lims filepath on the network to work on windows.

    Args:
        path: Lims filepath to 'fix'.

    Returns:
        The 'fix'ed path.
    """
    if path.startswith("/allen"):
        return "/" + path

    return path


def is_valid_exp_id(value: str) -> bool:
    """Checks whether or not an experiment id is valid.

    Args:
        value: Experiment id to validate.

    Returns:
        Is a valid experiment id.
    """
    return value.isdigit()



"""Awesome `np-lims-tk` is a Python cli/package created with https://github.com/TezRomacH/python-package-template"""

import sys

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata

from .core import find_files, get_suffixes, get_wkft_names


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

__all__ = [
    "find_files",
    "get_suffixes",
    "get_wkft_names",
    "version",
]



from typing import Any, Dict, List, Optional, Tuple, Union

import json
import logging
from dataclasses import replace

import yaml
from psycopg2 import connect, extensions, extras

from . import queries
from .config import allen_config_auto_discovery
from .dataclasses import Query
from .exceptions import LimsError, NPTKError
from .local_paths import get_suffixes, local_path_to_Query, path_to_lims_meta
from .utils import fix_lims_path, is_valid_exp_id, is_windows
from .wkft import get_wkft_names, wkft_to_Query

logger = logging.getLogger(__name__)


def run_query(cursor: extensions.cursor, query: Query) -> List[str]:
    """Runs a lims query.

    Args:
        cursor: lims postgresql cursor instance
        query: Query instance to run

    Returns:
        List of filepaths to assets on lims.
    """
    cursor.execute(query.query_str)
    filtered = []
    for row in cursor.fetchall():
        for (key, value) in query.filters:
            if row[key] != value:
                break
        else:
            filtered.append(row[query.return_name])
    return filtered


def _init_cursor(db_uri: str) -> extensions.cursor:
    """Initializes a connection to the lims database."""
    logger.info("Connecting to database uri: %s..." % db_uri)
    con = connect(db_uri)
    con.set_session(readonly=True, autocommit=True)
    return con.cursor(
        cursor_factory=extras.RealDictCursor,
    )


def _resolve_query(
    exp_id: Optional[str] = None,
    wkft: Optional[str] = None,
    local_path: Optional[str] = None,
) -> Query:
    """Resolves a combination of experiment meta info to a query to be executed on lims."""
    if wkft and exp_id:
        if not is_valid_exp_id(exp_id):
            raise LimsError("Invalid experiment id: %s" % exp_id)
        query = wkft_to_Query(wkft, exp_id)
    elif local_path:
        query = local_path_to_Query(local_path, exp_id)
    else:
        raise LimsError("Invalid number of arguments supplied.")

    return query


def find_files(
    db_uri: Optional[str] = None,
    exp_id: Optional[str] = None,
    wkft: Optional[str] = None,
    local_path: Optional[str] = None,
) -> List[str]:
    """Finds all files associated with a combination of experiment meta info.

    Args:
        db_uri: lims database uri
        exp_id: lims experiment id associated with the files
        wkft: well known file type name associated with the files you want to find
        local_path: local filepath of a file you want to find on lims

    Returns:
        A list of filepaths

    Note:
        - The following combinations of meta info are supported
            - ``exp_id`` and ``wkft``
            - ``local_path``
        - A ``local_path`` can be prefixed with * but an experiment id will need to be supplied
        - If no ``db_uri`` is supplied, will attempt ``allen_config_auto_discovery``
    """
    if db_uri is None:
        logger.info("No db_uri supplied, using allen config auto discovery...")
        db_uri = allen_config_auto_discovery("lims_db_uri")
        logger.info("Resolved db_uri: %s" % db_uri)

    query = _resolve_query(exp_id=exp_id, wkft=wkft, local_path=local_path)
    logger.info("Resolved query: %s" % query)

    with _init_cursor(db_uri) as cursor:
        results = run_query(cursor, query)

    if is_windows():
        logger.info(
            "Windows detected. Fixing paths for windows. Paths before: %s" % results
        )
        fixed = [fix_lims_path(result) for result in results]
        logger.info("Paths after: %s" % fixed)
        return fixed

    return results


def get_project_name(
    project_id: Optional[str] = None,
    lims_id: Optional[str] = None,
    path: Optional[str] = None,
    db_uri: Optional[str] = None,
) -> Union[str, None]:
    if not path and not lims_id and not project_id:
        raise Exception("Insufficient arguments.")
    if not lims_id and not project_id and path:
        meta = path_to_lims_meta(path)
        if not meta:
            raise NPTKError("Failed to deserialize meta info from path=%s" % path)

        lims_id = meta._id

    if db_uri is None:
        logger.info("No db_uri supplied, using allen config auto discovery...")
        db_uri = allen_config_auto_discovery("lims_db_uri")
        logger.info("Resolved db_uri: %s" % db_uri)

    with _init_cursor(db_uri) as cursor:
        if not project_id:
            results = run_query(
                cursor,
                Query(
                    queries.ECEPHYS_SESSION_QUERY.format(lims_id),
                    filters=[],
                    return_name="project_id",
                ),
            )
            project_id = results[0]

        results = run_query(
            cursor,
            Query(
                queries.PROJECT_QUERY.format(project_id), filters=[], return_name="code"
            ),
        )
        name = results[0]

    return name


def project_name_to_data_manifestor_template(
    name: str,
    manifestor_lookup: Optional[Dict[str, str]] = None,
    raw: bool = False,
) -> Union[str, Dict[str, Any]]:
    if not manifestor_lookup:
        manifestor_lookup = allen_config_auto_discovery(
            "manifestor_lookup",
        )

    try:
        path = manifestor_lookup[name]  # type: ignore
        if raw:
            return path

        with open(path) as f:
            return yaml.safe_load(f)  # type: ignore
    except KeyError:
        raise NPTKError("Invalid name=%s" % name)


def get_data_manifestor_template(
    lims_id: Optional[str] = None,
    manifestor_lookup: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    if not manifestor_lookup:
        manifestor_lookup = allen_config_auto_discovery(
            "manifestor_lookup",
        )

    project_name = get_project_name(lims_id=lims_id)

    try:
        path = manifestor_lookup[project_name]  # type: ignore
        if path.endswith(".yml"):
            with open(path) as f:
                return yaml.safe_load(f)  # type: ignore
        elif path.endswith(".json"):
            with open(path) as f:
                return json.load(f)  # type: ignore
        else:
            raise NPTKError("Unsupported data-manifestor file format: %s" % path)
    except KeyError:
        raise NPTKError("Invalid project_name=%s" % project_name)



from typing import Any

import json
import os

from .exceptions import NPTKError

ALLEN_CONFIG_PATH = (
    "//allen/programs/braintv/workgroups/nc-ophys/1022/allen_auto_config.json"
)


def allen_config_auto_discovery(key: str, config_path: str = ALLEN_CONFIG_PATH) -> Any:
    """For ease of development at the institute, gets config values
    from a static path on the /allen network drive.

    Args:
        key: name of the config value to grab
        config_path: config filepath

    Returns:
        config value for key

    Raises:
        NPTKError: Config at `config_path` couldn't be found or the requested `key` doesn't exist.
    """
    if not os.path.isfile(config_path):
        raise NPTKError("Couldn't find config: %s" % config_path)

    with open(config_path) as f:
        config = json.load(f)

    try:
        return config[key]
    except KeyError:
        raise NPTKError("%s not found in allen config" % key)



from typing import Dict, List, Optional, Tuple, Union

import logging
import os

from .dataclasses import LimsMeta, Query
from .exceptions import LimsError
from .wkft import wkft_to_Query

logger = logging.getLogger(__name__)

WILD_CARD = "*"

SUFFIX_TO_WKFT_MAP = {
    "_platformD1.json": "EcephysPlatformFile",
    ".sync": "EcephysRigSync",
    "_surface-image6-right.png": "EcephysPreInsertionRight",
    "_surface-image6-left.png": "EcephysPreInsertionLeft",
    "_surface-image5-left.png": "EcephysPreExperimentLeft",
    "_surface-image5-right.png": "EcephysPreExperimentRight",
    "_surface-image4-left.png": "EcephysPostStimulusLeft",
    "_surface-image4-right.png": "EcephysPostStimulusRight",
    "_surface-image3-right.png": "EcephysPostInsertionRight",
    "_surface-image3-left.png": "EcephysPostInsertionLeft",
    "_surface-image2-right.png": "EcephysPostExperimentRight",
    "_surface-image2-left.png": "EcephysPostExperimentLeft",
    ".overlay.png": "EcephysOverlayImage",
    ".insertionLocation.png": "EcephysInsertionLocationImage",
    ".fiducial.png": "EcephysFiducialImage",
    "_surface-image1-right.png": "EcephysBrainSurfaceRight",
    "_surface-image1-left.png": "EcephysBrainSurfaceLeft",
    ".behavior.pkl": "StimulusPickle",
    ".mapping.pkl": "MappingPickle",
    ".areaClassifications.csv": "EcephysAreaClassifications",
    ".behavior.json": "RawBehaviorTrackingVideoMetadata",
    ".behavior.mp4": "RawBehaviorTrackingVideo",
    ".eye.json": "RawEyeTrackingVideoMetadata",
    ".eye.mp4": "RawEyeTrackingVideo",
    ".face.json": "RawFaceTrackingVideoMetadata",
    ".face.mp4": "RawFaceTrackingVideo",
    ".motor-locs.csv": "NewstepConfiguration",
    ".replay.pkl": "EcephysReplayStimulus",
    ".opto.pkl": "OptoPickle",
    "_surgeryNotes.json": "EcephysSurgeryNotes",
}

SUFFIXES = list(SUFFIX_TO_WKFT_MAP.keys())
"""Names of supported suffixes for local_paths.
"""


def dir_to_experiment_id(path: str, suffix: str) -> str:
    """Currently unsupported"""
    raise LimsError("Directories not supported yet :0...")


def path_to_experiment_id(path: str, suffix: str) -> Union[str, None]:
    """Extracts an experiment id from a structured filepath.

    :param path: structured filepath
    :param suffix: suffix of the filepath
    :returns: extracted experiment id
    """
    if os.path.isdir(path):
        logger.info("Directory detected, using dir_to_meta.")
        meta = path_to_lims_meta(path)
        logger.debug("Meta: %s" % meta)
        if meta is not None:
            return meta._id
        else:
            return None

    filename = os.path.basename(path)
    meta_str = filename.removesuffix(suffix)

    if meta_str == WILD_CARD:
        logger.info("Wild card detected Not inferring exp id. path: %s" % path)
        return None

    values = meta_str.split("_")

    if len(values) != 3:  # assume valid meta_strs will be 3 items
        logger.info("Meta could not be parsed for: %s" % path)
        return None

    return values[0]


def deserialize_path(path: str) -> Tuple[Optional[str], str]:
    """Deserializes a structured filepath into an experiment id and
    wkft

    :param path: structured filepath
    :returns: experiment id if it could be extracted and wkft

    note:: wkft will be one of the well known file types supported by lims
    """
    for (
        suffix,
        wkft,
    ) in SUFFIX_TO_WKFT_MAP.items():
        if path.endswith(suffix):
            return (
                path_to_experiment_id(path, suffix),
                wkft,
            )
    else:
        raise LimsError("Unsupported path: %s" % path)


def local_path_to_Query(path: str, experiment_id: Optional[str] = None) -> Query:
    """Transforms a local filepath into a :class:`Query`.

    :param path: path to the file
    :param experiment_id: lims experiment id

    :returns: resolved query
    """
    inferred_experiment_id, wkft = deserialize_path(
        normalize_path(path),
    )

    if not experiment_id and not inferred_experiment_id:
        raise LimsError(
            "No experiment id supplied and no experiment id could be inferred. path: %s"
            % path
        )

    return wkft_to_Query(
        # raises an error above if both dont exist but maybe rethink
        wkft,
        experiment_id or inferred_experiment_id,  # type: ignore[arg-type]
    )  # choose explicit experiment id first


def normalize_path(path: str) -> str:
    """Transform a local path into parent_dir and filename."""
    if path.startswith(WILD_CARD):
        logger.info("Wild card detected, not normalizing. path: %s" % path)
        return path
    logger.info("Normalizing path: %s" % path)
    dirname, filename = path.split("\\")[-2:]
    normalized = os.path.join(dirname, filename)
    logger.info("Normalized path: %s" % normalized)
    return normalized


def get_suffixes() -> List[str]:
    """Gets a list of supported filename suffixes.

    Returns:
        list of filename suffixes
    """
    return list(SUFFIX_TO_WKFT_MAP.keys())


def path_to_lims_meta(path: str) -> Union[LimsMeta, None]:
    """Extracts serialized lims meta info from path."""
    try:
        base = os.path.basename(path.rstrip("/"))  # remove trailing / for dir
        _id, subject_id, date_str = (base.split(".")[0]).split("_")
    except ValueError:
        return None

    return LimsMeta(
        _id=_id,
        subject_id=subject_id,
        date_str=date_str,
    )



from typing import Optional

import json
from random import choice

import typer
from rich.console import Console

from np_lims_tk import core, exceptions, version

app = typer.Typer(
    name="np_lims_tk",
    help="Awesome `np_lims_tk` is a Python cli/package created with https://github.com/TezRomacH/python-package-template",
    add_completion=False,
)
console = Console()


def version_callback(print_version: bool) -> None:
    """Print the version of the package."""
    if print_version:
        console.print(f"[yellow]np-lims-tk[/] version: [bold blue]{version}[/]")
        raise typer.Exit()


@app.command(name="")
def main(
    print_version: bool = typer.Option(
        None,
        "-v",
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Prints the version of the np_lims_tk package.",
    ),
) -> None:
    """Print version."""


@app.command()
def find_lims_files(
    filepaths_json: str,
    output_path: str,
    db_uri: Optional[str] = typer.Option(
        None,
        "-d",
        "--db-uri",
        help="Database uri for lims connection. If not supplied, defaults to allen_config_auto_discovery.",
    ),
) -> None:
    with open(filepaths_json) as f:
        filepaths = json.load(f)

    results = []
    for filepath in filepaths:
        try:
            found = core.find_files(
                db_uri=db_uri,
                local_path=filepath,
            )
            error = None
        except Exception as e:
            found = None
            error = str(e)

        result = {
            "filepath": filepath,
            "found": found,
            "error": error,
        }
        results.append(result)

    with open(output_path, "w") as f:
        json.dump(
            results,
            f,
            sort_keys=True,
            indent=4,
        )


@app.command()
def path_to_data_manifestor_meta(
    path: str,
    output_path: str,
    db_uri: Optional[str] = typer.Option(
        None,
        "-d",
        "--db-uri",
        help="Database uri for lims connection. If not supplied, defaults to allen_config_auto_discovery.",
    ),
    manifestor_lookup_path: Optional[str] = typer.Option(
        None,
    ),
) -> None:
    if manifestor_lookup_path:
        with open(manifestor_lookup_path) as f:
            manifestor_lookup = json.load(f)
    else:
        manifestor_lookup = None

    lims_meta = core.path_to_lims_meta(path)
    if not lims_meta:
        raise exceptions.NPTKError("No lims meta inferred from path: %s" % path)

    lims_id = lims_meta._id
    project_name = core.get_project_name(
        lims_id=lims_id,
        db_uri=db_uri,
    )

    if not project_name:
        raise exceptions.NPTKError(
            "Couldnt find associated project name. lims_id={}, db_uri={}".format(
                lims_id, db_uri
            )
        )

    template_path = core.project_name_to_data_manifestor_template(
        project_name,
        manifestor_lookup=manifestor_lookup,
        raw=True,
    )

    meta = {
        "lims_id": lims_id,
        "subject_id": lims_meta.subject_id,
        "date_str": lims_meta.date_str,
        "project_name": project_name,
        "manifestor_template_path": template_path,
    }

    with open(output_path, "w") as f:
        json.dump(meta, f)


if __name__ == "__main__":
    app()



EXP_WKFT_QUERY = """
SELECT es.id AS es_id, 
    es.name AS es,
    es.storage_directory,
    es.workflow_state,
    es.date_of_acquisition,
    es.stimulus_name,
    es.foraging_id as foraging_id,
    sp.external_specimen_name,
    isi.id AS isi_experiment_id,
    e.name AS rig,
    u.login AS operator,
    p.code AS project,
    wkft.name AS wkft, 
    wkf.storage_directory || wkf.filename AS wkf_path,
    bs.storage_directory AS behavior_dir
FROM ecephys_sessions es
    JOIN specimens sp ON sp.id = es.specimen_id
    LEFT JOIN isi_experiments isi ON isi.id = es.isi_experiment_id
    LEFT JOIN equipment e ON e.id = es.equipment_id
    LEFT JOIN users u ON u.id = es.operator_id
    JOIN projects p ON p.id = es.project_id
    LEFT JOIN well_known_files wkf ON wkf.attachable_id = es.id
    LEFT JOIN well_known_file_types wkft ON wkft.id=wkf.well_known_file_type_id
    LEFT JOIN behavior_sessions bs ON bs.foraging_id = es.foraging_id
WHERE es.id = {}
ORDER BY es.id
"""

IMAGE_WKFT_QUERY = """
SELECT es.id AS es_id, es.name AS es, imt.name AS image_type, es.storage_directory || im.jp2 AS image_path
FROM ecephys_sessions es
    JOIN observatory_associated_data oad ON oad.observatory_record_id = es.id AND oad.observatory_record_type = 'EcephysSession'
    JOIN images im ON im.id=oad.observatory_file_id AND oad.observatory_file_type = 'Image'
    JOIN image_types imt ON imt.id=im.image_type_id
WHERE es.id = {}
ORDER BY es.id, imt.name;
"""

PROBE_WKFT_QUERY = """
SELECT es.id AS es_id, 
    es.name AS es, 
    ep.name AS ep, 
    ep.id AS ep_id, 
    wkft.name AS wkft, 
    wkf.storage_directory || wkf.filename AS wkf_path
FROM ecephys_sessions es
    JOIN ecephys_probes ep ON ep.ecephys_session_id=es.id
    LEFT JOIN well_known_files wkf ON wkf.attachable_id = ep.id
    LEFT JOIN well_known_file_types wkft ON wkft.id=wkf.well_known_file_type_id
WHERE es.id = {} 
ORDER BY es.id, ep.name;
"""

ECEPHYS_SESSION_QUERY = """
SELECT *
FROM ecephys_sessions es
WHERE es.id = {}
"""

PROJECT_QUERY = """
SELECT *
FROM projects p
WHERE p.id={}
"""



from typing import Any, List, Optional, Tuple

from dataclasses import dataclass


@dataclass
class Query:

    """Abstraction of query to be executed on lims database."""

    query_str: str
    filters: List[Tuple[str, Any]]
    return_name: str


@dataclass
class LimsMeta:

    """Abstraction of np project metadata."""

    _id: str
    subject_id: str
    date_str: str
